import os
from collections import defaultdict


def find_tiles(tiledir):
    for xdir in os.listdir(tiledir):
        try:
            x = int(xdir)
        except ValueError:
            continue
        for filename in os.listdir(os.path.join(tiledir, xdir)):
            base, ext = os.path.splitext(filename)
            y = int(base)
            yield (x, y)


def make_pyramides(basedir, level, tiles):
    if level == 0:
        return
    tiles = set(tiles)
    next_tiles = set((int(x/2), int(y/2)) for x, y in tiles)

    def intile_name(x, y):
        if (x, y) in tiles:
            return os.path.join(basedir, str(level), str(x), "%d.png" % y)
        else:
            return "-"

    def outtile_name(x, y):
        return os.path.join(basedir, str(level-1), str(x), "%d.png" % y)

    for x, y in next_tiles:
        innames = [intile_name(2*x+sx, 2*y+sy)
                   for sx, sy in [(0, 0), (0, 1), (1, 0), (1, 1)]]
        indeps = [inname for inname in innames if inname != "-"]
        print(outtile_name(x, y) + ": " + (" ".join(indeps)))
        print("\t${PYRAMID_STEP} $@ " + (" ".join(innames)))
        print()
    make_pyramides(basedir, level-1, next_tiles)


def _main(from_setuptools_script=True):
    import sys
    if from_setuptools_script:
        pyramid_step = "pyramid_step"
        overlay_tiles = "overlay_tiles"
    else:
        script_base_dir = os.path.abspath(os.path.dirname(__file__))
        pyramid_step = "\"{}\" \"{}\"".format(
                sys.executable,
                os.path.join(script_base_dir, "pyramid_step.py"))
        overlay_tiles = "\"{}\" \"{}\"".format(
                sys.executable,
                os.path.join(script_base_dir, "overlay_tiles.py"))

    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("base_level",
                        type=int,
                        help="level in which input tiles have been generated")
    parser.add_argument("base_dir",
                        type=str,
                        help="folder containing input tiles")
    parser.add_argument("out_dir",
                        type=str,
                        help="folder for generated output tiles")
    args = parser.parse_args()

    base_level = args.base_level
    basedir = args.base_dir
    outdir = args.out_dir
    layers = sorted(os.listdir(basedir))
    layers_by_tile = defaultdict(list)
    for layer in layers:
        for tile in find_tiles(os.path.join(basedir, layer, str(base_level))):
            layers_by_tile[tile].append(layer)

    def outname(x, y):
        return os.path.join(outdir,
                            str(base_level), str(x), str(y) + ".png")

    def inname(layer, x, y):
        return os.path.join(basedir, layer,
                            str(base_level), str(x), str(y) + ".png")

    yvals_by_x = defaultdict(list)
    for x, y in layers_by_tile.keys():
        yvals_by_x[x].append(y)

    print("PYRAMID_STEP={}".format(pyramid_step))
    print("OVERLAY_TILES={}".format(overlay_tiles))
    print()
    print("all: " + os.path.join(outdir, "0", "0", "0.png"))
    print(".PHONY: all")
    print()
    make_pyramides(outdir, base_level, layers_by_tile.keys())
    for (x, y), layers in sorted(layers_by_tile.items()):
        print(outname(x, y) + ": " + (" ".join(inname(layer, x, y)
                                               for layer in sorted(layers))))
        print("\t${OVERLAY_TILES} $@ $^")
        print()


if __name__ == '__main__':
    _main(False)



from PIL import Image

TILE_SIZE = 256


def combine(tl, bl, tr, br):
    new_image = Image.new("RGBA", (2*TILE_SIZE, 2*TILE_SIZE), (0, 0, 0, 0))
    for xofs, yofs, subtile in [(0, 0, tl),
                                (1, 0, tr),
                                (0, 1, bl),
                                (1, 1, br)]:
        new_image.paste(subtile,
                        box=(xofs*TILE_SIZE, yofs*TILE_SIZE,
                             (xofs+1)*TILE_SIZE, (yofs+1)*TILE_SIZE))
    new_image.thumbnail((TILE_SIZE, TILE_SIZE), Image.LANCZOS)
    return new_image


def load_file(path):
    if path == "-":
        return Image.new("RGBA", (TILE_SIZE, TILE_SIZE), (0, 0, 0, 0))
    else:
        return Image.open(path)


def _main():
    import os
    import sys
    outfile = sys.argv[1]
    infiles = sys.argv[2:]
    assert len(infiles) == 4
    outdir = os.path.dirname(outfile)
    if not os.path.exists(outdir):
        try:
            os.makedirs(outdir)
        except OSError:
            # catch race condition, if folder is created in between both calls
            pass
    outimage = combine(*map(load_file, infiles))
    outimage.save(outfile)


if __name__ == '__main__':
    _main()



from PIL import Image


def overlay(images):
    if len(images) == 1:
        return images[0]
    new_image = Image.new("RGBA", images[0].size, (0, 0, 0, 0))
    for image in images:
        new_image.paste(image, None, image)
    return new_image


def _main():
    import os
    import sys
    outfile = sys.argv[1]
    infiles = sys.argv[2:]
    outdir = os.path.dirname(outfile)
    if not os.path.exists(outdir):
        try:
            os.makedirs(outdir)
        except OSError:
            # catch race condition, if folder is created in between both calls
            pass
    new_im = overlay(list(map(Image.open, infiles)))
    new_im.save(outfile)


if __name__ == '__main__':
    _main()



import os
import itertools
import numpy as np
from PIL import Image


def numTiles(z):
    return 2**z


def resolution2zoom(res, lat=0., tilesize=256):
    """
    computes zoom level corresponsing to a given desired resolution
    :param res: desired resolution in meters
    :param lat: lowest absolute latitude value of desired scene
    :param tilesize: size of one tile in pixels

    :see: https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Resolution_and_Scale # noqa: E501
    """
    zoom0_pixel_size = 40075016.686 / tilesize
    return np.log2(zoom0_pixel_size * np.cos(np.deg2rad(lat)) / res)


def latlon2relativeXY(lat, lon):
    x = (lon + 180) / 360
    rlat = np.deg2rad(lat)
    y = (1 - np.log(np.tan(rlat) + 1./np.cos(rlat)) / np.pi) / 2
    return x, y


def latlon2xy(lat, lon, z):
    n = numTiles(z)
    x, y = latlon2relativeXY(lat, lon)
    return n*x, n*y


def mercatorToLat(mercatorY):
    return np.rad2deg(np.arctan(np.sinh(mercatorY)))


def xy2latlon(x, y, z):
    n = numTiles(z)
    rel_y = y / n
    lat = mercatorToLat(np.pi * (1 - 2 * rel_y))
    lon = -180.0 + 360.0 * x / n
    return lat, lon


def render_tile(x, y, z, callback, tilesize=256):
    ys = ((np.arange(tilesize) + .5) / tilesize)[:, np.newaxis]
    xs = ys.T
    xs, ys = np.broadcast_arrays(x + xs, y + ys)
    lat, lon = xy2latlon(xs, ys, z)
    return callback(lat, lon)


class LLTiler:
    def __init__(self,
                 data_folder,
                 base_level=None,
                 size_hint=None,
                 naming_scheme="{z}/{x}/{y}.png"):
        """
        :param data_folder: folder to store tiles
        :param base_level: level in which tiles should be computed
        :param size_tint: smallest size meters to be resolved
        :param naming_scheme: pattern for tilenames,
                              must include x, y and z placeholders

        :note: either `base_level` or `size_hint` must be given
        """
        self.data_folder = data_folder
        if base_level is None:
            if size_hint is None:
                raise ValueError(
                        "either base_level or size_hint must be given")
        self.size_hint = size_hint
        self.base_level = base_level

        self.naming_scheme = naming_scheme

    def render(self, extent, callback, show_progress=False):
        """
        :param extent: ((lat_min, lon_min), (lat_max, lon_max))
        :param callback: function accepting lat and a lon array
                         and returns RGBA values
        """

        if show_progress:
            from tqdm import tqdm
            progress = tqdm
        else:
            def progress(x, total):
                return x

        (lat_min, lon_min), (lat_max, lon_max) = extent

        if lat_min * lat_max < 0:
            lat_abs_min = 0.
        else:
            lat_abs_min = min(abs(lat_min), abs(lat_max))

        if self.base_level is None:
            self.base_level = int(np.ceil(resolution2zoom(
                self.size_hint, lat_abs_min)))

        x1, y1 = latlon2xy(lat_min, lon_min, self.base_level)
        x2, y2 = latlon2xy(lat_max, lon_max, self.base_level)

        x_min = int(min(x1, x2))
        x_max = int(max(x1, x2))
        y_min = int(min(y1, y2))
        y_max = int(max(y1, y2))

        for x, y in progress(itertools.product(range(x_min, x_max + 1),
                                               range(y_min, y_max + 1)),
                             total=(x_max - x_min + 1) * (y_max - y_min + 1)):
            tile = render_tile(x, y, self.base_level, callback)
            self.store_tile(tile, x, y, self.base_level)

    def tile_path(self, x, y, z):
        return os.path.join(self.data_folder,
                            self.naming_scheme.format(x=x, y=y, z=z))

    def store_tile(self, tile, x, y, z):
        filename = self.tile_path(x, y, z)
        folder = os.path.dirname(filename)
        if not os.path.exists(folder):
            os.makedirs(folder)
        image = Image.fromarray(tile)
        image.save(filename)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lltiler",
    version="0.0.1",
    author="Tobias Kölling",
    author_email="tobias.koelling@physik.uni-muenchen.de",
    description=(
        "The Lat-Lon Tiler is a small tool to create map tiles in spherical"
        "mercator projection for the use with online map libraries like"
        "Leaflet, OpenLayers, Mapbox etc..."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/d70-t/lltiler",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
    install_requires=[
        "numpy",
        "pillow",
    ],
    entry_points={
        "console_scripts": [
            "{0} = lltiler.scripts.{0}:_main".format(script)
            for script in ["overlay_tiles",
                           "pyramid_step",
                           "generate_tile_makefile"]
        ],
    },
)
